﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;


namespace GetTheHeightOfTheTerrainModel
{
    // 高さ情報クラス
    public class HeightInfo
    {
        // 地形モデルの頂点数（幅と高さ）を指定すること
        private const int vertexWidth = 65;
        private const int vertexheight = 65;


        // カスタム頂点フォーマット構造体の宣言
        // 注意：地形モデルの頂点フォーマットに合わせて変更すること
        private struct VertexPositionTextureNormal
        {
            public Vector3 Position;
            public Vector2 TextureCoordinate;
            public Vector3 Normal;

            public VertexPositionTextureNormal(Vector3 position, Vector2 textureCoordinate, Vector3 normal)
            {
                this.Position = position;
                this.TextureCoordinate = textureCoordinate;
                this.Normal = normal;
            }

            public static readonly VertexElement[] VertexElements
                = {
                new VertexElement(
                    0,
                    0,
                    VertexElementFormat.Vector3,
                    VertexElementMethod.Default,
                    VertexElementUsage.Position,
                    0
                ),

                new VertexElement(
                    0,
                    sizeof(float) * 3,
                    VertexElementFormat.Vector2,
                    VertexElementMethod.Default,
                    VertexElementUsage.TextureCoordinate,
                    0
                ),

                new VertexElement(
                    0, 
                    sizeof(float) * (3 + 2),
                    VertexElementFormat.Vector3,
                    VertexElementMethod.Default,
                    VertexElementUsage.Normal,
                    0
                )
            };
        }



        // 地形のスケールの宣言
        private float terrainScale;

        // 高さ情報配列の宣言
        private float[,] heights;

        // 地形の座標（左上を基準とするための差分）の宣言
        private Vector3 heightmapPosition;

        // 地形の幅の宣言
        private float heightmapWidth;
        public float HeightmapWidth
        {
            get
            {
                return heightmapWidth;
            }
        }

        // 地形の高さの宣言
        private float heightmapHeight;
        public float HeightmapHeight
        {
            get
            {
                return heightmapHeight;
            }
        }



        // コンストラクタ
        public HeightInfo(Model map, float mapScale)
        {
            // 地形モデルから頂点データのコピーを取得する
            int elementCount = vertexWidth * vertexheight;
            VertexPositionTextureNormal[] v = new VertexPositionTextureNormal[elementCount];
            BufferUsage bu = map.Meshes[0].VertexBuffer.BufferUsage;
            map.Meshes[0].VertexBuffer.GetData<VertexPositionTextureNormal>(v, 0, elementCount);

            // 座標値をソート用配列にコピーする
            Vector3[] pos = new Vector3[elementCount];
            for (int i = 0; i < elementCount; i++)
                pos[i] = v[i].Position;

            // バブルソート
            // Ｃ＃のArray.Sortに安定性は無い
            // （比較の結果が同一順位だった場合に、並べ替え項目のもともとの位置関係が維持されない）
            // ので、二段階ソートができない。よってバブルソートを使用した。（マージソートなどでも可）
            bubbleSortZ(pos);
            bubbleSortX(pos);

            // 同一の頂点座標データを除去する
            // 参考：CGツールで表示される頂点数と、XNAでモデルを読み込んだ際の頂点数は異なることがある。
            // 参考：CGツールの地形作成機能のみ使用したモデルでも、頂点数が幅×高さより多いことがある。
            List<Vector3> newPos = removeSameValue(pos);

            // 頂点数をチェック
            if (newPos.Count != vertexWidth * vertexheight)
                throw new ArgumentException("頂点数が一致しません！");

            // terrainScaleを算出
            float terrainScale = ((newPos[vertexWidth * vertexheight - 1].X - newPos[0].X) / ((float)vertexWidth - 1.0f)) * mapScale;
            this.terrainScale = terrainScale;

            // heightScaleを算出
            float heightScale = ((float)vertexWidth - 1.0f) / (newPos[vertexWidth * vertexheight - 1].X - newPos[0].X);

            // 二次元配列に代入する
            this.heights = new float[vertexWidth, vertexheight];
            for (int x = 0; x < vertexWidth; x++)
            {
                for (int z = 0; z < vertexheight; z++)
                {
                    // 高さ情報にスケールを反映
                    this.heights[x, z] = newPos[x * vertexheight + z].Y * heightScale * terrainScale;
                }
            }

            // 地形の幅と高さを設定
            heightmapWidth = (heights.GetLength(0) - 1) * terrainScale;
            heightmapHeight = (heights.GetLength(1) - 1) * terrainScale;

            // 地形の座標（左上を基準とするための差分）を設定
            heightmapPosition.X = -(heights.GetLength(0) - 1) / 2 * terrainScale;
            heightmapPosition.Z = -(heights.GetLength(1) - 1) / 2 * terrainScale;
        }


        // X座標でバブルソート
        private void bubbleSortX(Vector3[] pos)
        {
            int n = pos.Length;
            for (int i = 0; i < n - 1; i++)
                for (int j = n - 1; j > i; j--)
                    if (pos[j].X < pos[j - 1].X)
                        swap(ref pos[j], ref pos[j - 1]);
        }

        // Z座標でバブルソート
        private void bubbleSortZ(Vector3[] pos)
        {
            int n = pos.Length;
            for (int i = 0; i < n - 1; i++)
                for (int j = n - 1; j > i; j--)
                    if (pos[j].Z < pos[j - 1].Z)
                        swap(ref pos[j], ref pos[j - 1]);
        }

        // a と b の中身を入れ替える（バブルソート用）
        private void swap<Vector3>(ref Vector3 pos1, ref Vector3 pos2)
        {
            Vector3 tmp = pos1;
            pos1 = pos2;
            pos2 = tmp;
        }


        // 同一の頂点座標データを除去する
        public static List<Vector3> removeSameValue(Vector3[] pos)
        {
            // 返却用リストの宣言
            List<Vector3> retList = new List<Vector3>();
            // 最初の値を代入
            retList.Add(pos[0]);
            for (int i = 1; i < pos.Length; i++)
                // 異なる値か？
                if (pos[i] != pos[i - 1])
                    // 異なる値なら代入する
                    retList.Add(pos[i]);
            return retList;
        }



        // この関数は位置を受け取り、その位置がハイトマップ上にあるかどうかを判別します。
        public bool IsOnHeightmap(Vector3 position)
        {
            // 最初に、ハイトマップ上のどこに "positionOnHeightmap" があるかを算出し...
            Vector3 positionOnHeightmap = position - heightmapPosition;

            // 次に、その値がハイトマップの境界外になるかどうかを調べます。
            return (positionOnHeightmap.X > 0 &&
                positionOnHeightmap.X < heightmapWidth &&
                positionOnHeightmap.Z > 0 &&
                positionOnHeightmap.Z < heightmapHeight);
        }


        // この関数は、位置を受け取り、そのポイントでのハイトマップの高さを返します。
        // http://create.msdn.com/ja-JP/education/catalog/sample/collision_3d_heightmap
        //
        // この関数は、上記の公式サンプルにメタセコイア地形モデル用の改修を加えた
        // 
        public float GetHeight(Vector3 position)
        {
            // 最初に行う必要のあることは、ハイトマップ上のどこに "position" があるかを
            // 算出することです。これにより、後の計算をより単純化することができます。
            Vector3 positionOnHeightmap = position - heightmapPosition;

            // 整数除算を使用して、"heights" 配列のどこに positionOnHeightmap が
            // あるかを算出します。整数除算の端数が常に切り捨てられ、これらの除算の結果が、
            // そのセルの 4 つのコーナーの "左上" を示すインデックスになることに
            // 留意してください。
            int left, top;
            left = (int)positionOnHeightmap.X / (int)terrainScale;
            top = (int)positionOnHeightmap.Z / (int)terrainScale;



            // 位置がハイトマップ上にない場合に発生するエラーを回避するための補正
            int leftAdd = 1;
            int topAdd = 1;

            if (left < 0)
                left = 0;
            if (left >= heights.GetLength(1) - 1)
            {
                left = heights.GetLength(1) - 1;
                leftAdd = 0;
            }

            if (top < 0)
                top = 0;
            if (top >= heights.GetLength(0) - 1)
            {
                top = heights.GetLength(0) - 1;
                topAdd = 0;
            }


            // 次に、剰余演算を使用して、セルの左上コーナーからどれだけ離れているかを
            // 算出します。剰余演算により、0 から terrainScale までの値が返されます。
            // 続いてこの値を terrainScale で除算して、0 から 1 に正規化します。
            float xNormalized = (positionOnHeightmap.X % terrainScale) / terrainScale;
            float zNormalized = (positionOnHeightmap.Z % terrainScale) / terrainScale;

            // セルのコーナーのインデックスとそのセル内での位置を計算したので、バイリニア
            // 補間を使用して高さを計算します。このプロセスは図を参照したほうが理解しやすいので、
            // 付属のドキュメントを参照してください。
            // 最初に、左側と右側から補間することによって、セルの上端と下端での高さを計算します。

            // メタセコイア地形モデルは、１つのセルが２つの三角形で構成され、
            // 各面の角度が異なるので、地形形状がXNA地形モデルと異なる。

            float topHeight;
            float bottomHeight;
            if (xNormalized > zNormalized)
            {
                // 三角形の左上→右上
                topHeight = MathHelper.Lerp(
                    heights[left, top],
                    heights[left + leftAdd, top],
                    xNormalized);

                // 三角形の左上→右下
                bottomHeight = MathHelper.Lerp(
                    heights[left, top],
                    heights[left + leftAdd, top + topAdd],
                    xNormalized);

                // zNormalizedを修正
                zNormalized = zNormalized / xNormalized;
            }
            else
            {
                // 三角形の左上→右下
                topHeight = MathHelper.Lerp(
                    heights[left, top],
                    heights[left + leftAdd, top + topAdd],
                    xNormalized);
                // 三角形の左下→右下
                bottomHeight = MathHelper.Lerp(
                    heights[left, top + topAdd],
                    heights[left + leftAdd, top + topAdd],
                    xNormalized);

                // zNormalizedを修正
                zNormalized = (zNormalized - xNormalized) / (1 - xNormalized);
            }

            // 次に、これら 2 つの値を線形補間して、位置での高さを計算します。
            return MathHelper.Lerp(topHeight, bottomHeight, zNormalized);
        }
    }
}
